#!/usr/bin/env python
"""
Module GROUPING -- Simple Blog Grouping Extension
Sub-Package SIMPLEBLOG.EXTENSIONS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from itertools import groupby
from operator import attrgetter

from plib.stdlib.decotools import cached_property

from simpleblog import extendable_property, BlogEntry
from simpleblog.extensions import BlogExtension


class GroupingPageMixin(object):
    
    @cached_property
    def groupindex_key(self):
        return self.blog.config.get('groupindex_key', 'groupindex')
    
    @cached_property
    def group_key(self):
        return self.blog.config.get('group_key', 'datestamp_formatted')
    
    @extendable_property()
    def group_head_template(self):
        return self.template_data("date", "head")
    
    @extendable_property()
    def group_foot_template(self):
        return self.template_data("date", "foot")
    
    @cached_property
    def group_formats(self):
        return set(self.config.get('group_formats', ["html"]))
    
    def _get_format_entries(self):
        if self.format in self.group_formats:
            for groupindex, (groupkey, group) in enumerate(groupby(
                self.entries, attrgetter(self.group_key)
            )):
                params = {
                    self.groupindex_key: groupindex,
                    self.group_key: groupkey
                }
                yield self.group_head_template.format(**params)
                group_params = dict(
                    groupindex=groupindex,
                    groupkey=groupkey
                )
                for entry in group:
                    entryparams = self.entry_params(entry)
                    entryparams.update(group_params)
                    yield entry.formatted(self.format, entryparams)
                yield self.group_foot_template.format(**params)
                first_group = False
        else:
            for item in super(GroupingPageMixin, self)._get_format_entries():
                yield item


class GroupingExtension(BlogExtension):
    """Group page entries by date.
    """
    
    page_mixin = GroupingPageMixin



#!/usr/bin/env python
"""
Module ARCHIVES -- Simple Blog Archives Extension
Sub-Package SIMPLEBLOG.EXTENSIONS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from collections import defaultdict

from plib.stdlib.decotools import cached_property

from simpleblog import BlogEntries, monthnames
from simpleblog.extensions import BlogExtension, get_links


year_entries = defaultdict(list)
month_entries = defaultdict(list)
day_entries = defaultdict(list)


class BlogArchiveEntries(BlogEntries):
    """Archive of entries spanning a given time period.
    """
    
    sourcetype = 'archive'
    
    def __init__(self, blog, year, month=0, day=0):
        BlogEntries.__init__(self, blog)
        self.year = year
        self.month = month
        self.monthname = monthname = monthnames[month]
        self.monthkey = monthkey = (
            "{:02d}".format(month), monthname)[self.archive_use_monthnames]
        self.day = day
        
        if month:
            if day:
                self.sortkey = (year, month, day)
                self.urlshort = "/{0}/{1}/{2}/".format(year, monthkey, day)
                title_template = self.archive_day_template
            else:
                self.sortkey = (year, month)
                self.urlshort = "/{0}/{1}/".format(year, monthkey)
                title_template = self.archive_month_template
        else:
            self.sortkey = (year,)
            self.urlshort = "/{}/".format(year)
            title_template = self.archive_year_template
        
        if self.prefix:
            self.urlshort = "/{0}{1}".format(self.prefix, self.urlshort)
        
        self.title = title_template.format(**vars(self))
        self.heading = "Archive: {}".format(self.title)
    
    @cached_property
    def prefix(self):
        return self.config.get('archives_prefix', "")
    
    @cached_property
    def archive_use_monthnames(self):
        return self.config.get('archive_use_monthnames', False)
    
    @cached_property
    def archive_year_template(self):
        return self.config.get('archive_year_template', "{year}")
    
    @cached_property
    def archive_month_template(self):
        return self.config.get('archive_month_template', "{year}-{monthkey}")
    
    @cached_property
    def archive_day_template(self):
        return self.config.get('archive_day_template', "{year}-{monthkey}-{day}")
    
    def _get_entries(self):
        if self.month:
            if self.day:
                return day_entries[(self.year, self.month, self.day)]
            return month_entries[(self.year, self.month)]
        return year_entries[self.year]


class ArchivesExtension(BlogExtension):
    """Add blog archive pages.
    """
    
    @cached_property
    def archive_years(self):
        return self.config.get('archive_years', False)
    
    @cached_property
    def archive_months(self):
        return self.config.get('archive_months', False)
    
    @cached_property
    def archive_days(self):
        return self.config.get('archive_days', False)
    
    @cached_property
    def archive_link_years(self):
        return self.config.get('archive_link_years', False)
    
    @cached_property
    def archive_link_months(self):
        return self.config.get('archive_link_months', False)
    
    @cached_property
    def archive_link_days(self):
        return self.config.get('archive_link_days', False)
    
    def blog_mod_sources(self, blog, sources):
        
        def archive_containers(t):
            if self.archive_years:
                yield year_entries[t.year]
            if self.archive_months:
                yield month_entries[(t.year, t.month)]
            if self.archive_days:
                yield day_entries[(t.year, t.month, t.day)]
        
        for entry in blog.all_entries:
            for container in archive_containers(entry.timestamp):
                container.append(entry)
        
        blog.all_archives = []
        archive_links = []
        
        if self.archive_years:
            years = [
                BlogArchiveEntries(blog, year)
                for year in year_entries
            ]
            blog.all_archives.extend(years)
            if self.archive_link_years:
                archive_links.extend(years)
        
        if self.archive_months:
            months = [
                BlogArchiveEntries(blog, year, month)
                for year, month in month_entries
            ]
            blog.all_archives.extend(months)
            if self.archive_link_months:
                archive_links.extend(months)
        
        if self.archive_days:
            days = [
                BlogArchiveEntries(blog, year, month, day)
                for year, month, day in day_entries
            ]
            blog.all_archives.extend(days)
            if self.archive_link_days:
                archive_links.extend(days)
        
        blog.metadata['archive_links'] = get_links(archive_links, True)
        
        sources.extend(
            (archive, "html")
            for archive in blog.all_archives
        )
        
        return sources



#!/usr/bin/env python
"""
Sub-Package SIMPLEBLOG.EXTENSIONS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from operator import attrgetter

from plib.stdlib.decotools import cached_property, wraps_class
from plib.stdlib.iters import prefixed_items

from simpleblog import (
    extension_types, extension_map, extend_attributes,
    BlogEntries, html_newline)


class NamedEntries(BlogEntries):
    """Named container for a set of blog entries.
    """
    
    typename = None
    prefix = None
    
    def __init__(self, blog, name):
        BlogEntries.__init__(self, blog)
        self.name = self.title = self.sortkey = name
        self.heading = "{0}: {1}".format(self.typename, name)
        if self.prefix:
            self.urlshort = "/{0}/{1}/".format(self.prefix, name)
        else:
            self.urlshort = "/{}/".format(name)
    
    @cached_property
    def prev_next_suffix(self):
        return self.name


def get_links(containers, reverse=False):
    """Return HTML links to containers.
    """
    # FIXME make this configurable
    return html_newline.join(
        '<a href="{0}">{1}</a>&nbsp;({2})'.format(c.urlshort, c.title, len(c.entries))
        for c in sorted(containers, key=attrgetter('sortkey'), reverse=reverse)
    )


class BlogExtension(object):
    """Base class for extension mechanism.
    """
    
    def __init__(self, config):
        self.config = config
        
        # Register this extension in the appropriate places
        attr_tmpl = '{}_'
        mixin_tmpl = '{}_mixin'
        for etype in extension_types:
            # Check for methods that extend a known attribute
            if any(
                item for item in prefixed_items(dir(self), attr_tmpl.format(etype))
                if not item.endswith('mixin')
            ):
                extension_map[etype].append(self)
            # Check for extensions that declare mixins
            mixin = getattr(self, mixin_tmpl.format(etype), None)
            if mixin:
                mixin.extension_type = etype
                extend_attributes(mixin)
                oldcls = extension_types[etype]
                @wraps_class(oldcls)
                class Extended(mixin, oldcls):
                    pass
                extension_types[etype] = Extended
        
        # Allow for post-init processing in subclasses
        self.post_init()
    
    def post_init(self):
        pass



#!/usr/bin/env python
"""
Module RENDER_STATIC -- Simple Blog Static Rendering
Sub-Package SIMPLEBLOG.COMMANDS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import os

from simpleblog.commands import BlogCommand


class RenderStatic(BlogCommand):
    """Static rendering of all blog pages.
    """
    
    def run(self, config, blog):
        for page in blog.pages:
            data = page.formatted
            path = os.path.abspath("{0}{1}".format(
                config.get('static_dir', "static"), page.urlpath
            ))  # FIXME make this portable
            print "Rendering", path
            dir = os.path.split(path)[0]
            if not os.path.isdir(dir):
                os.makedirs(dir)
            with open(path, 'w') as f:
                f.write(data)



#!/usr/bin/env python
"""
Module LOCALIZE -- Simple Blog Localization Extension
Sub-Package SIMPLEBLOG.EXTENSIONS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from simpleblog.extensions import BlogExtension


class LocalizeExtension(BlogExtension):
    """Allow config to specify localization for blog.
    """
    
    def post_init(self):
        config = self.config
        try:
            config.settings.setdefault('locale', "{0}_{1}.{2}".format(
                config.language, config.country, config.charset
            ))
        except AttributeError:
            pass



#!/usr/bin/env python
"""
Module TIMESTAMPS -- Simple Blog Timestamps Extension
Sub-Package SIMPLEBLOG.EXTENSIONS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import os
from datetime import datetime
from functools import total_ordering

from plib.stdlib.decotools import convert

from simpleblog import weekdaynames, monthnames, extendable_property
from simpleblog.caching import cached
from simpleblog.extensions import BlogExtension


@total_ordering
class Timestamp(object):
    """Blog entry time stamp.
    
    Wraps ``datetime.datetime`` to customize for usage here. Note that
    we use "naive" ``datetime`` objects; we don't expect blog entry
    times to be close enough to make correct handling of time zones,
    DST transitions, etc. to be an issue. This does, however, mean that
    "original" timestamps, generated from the entry source file mtime
    when the entry is first added to the timestamp file, are generated
    using platform local time, since we use ``datetime.fromtimestamp``
    to construct them.
    """
    
    datetime_attrs = (
        'year', 'month', 'day',
        'hour', 'minute', 'second',
        'weekday'
    )
    
    def __init__(self, data):
        if isinstance(data, float):
            # It's a file mtime
            self._datetime = datetime.fromtimestamp(data)
        elif isinstance(data, basestring):
            # It's a string timestamp from the cache
            self._datetime = datetime(*map(int, data.split('-')))
        else:
            raise ValueError("{!r} is not a valid entry timestamp".format(data))
        for attr in self.datetime_attrs:
            setattr(self, attr, getattr(self._datetime, attr))
    
    def __repr__(self):
        return "Timestamp: {!r}".format(self._datetime)
    
    def __str__(self):
        dt = self._datetime
        return '-'.join(
            '{:02d}'.format(i)
            for i in (dt.year, dt.month, dt.day, dt.hour, dt.minute)
        )
    
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._datetime == other._datetime
        if isinstance(other, datetime):
            return self._datetime == other
        raise NotImplementedError
    
    def __gt__(self, other):
        if isinstance(other, self.__class__):
            return self._datetime > other._datetime
        if isinstance(other, datetime):
            return self._datetime > other
        raise NotImplementedError


timestamps_file = BlogExtension.config.get('timestamps_file', "timestamps")


class TimestampEntryMixin(object):
    
    @extendable_property(
        cached(timestamps_file, reverse=True, objtype=Timestamp)
    )
    def timestamp(self):
        return os.path.getmtime(self.filename)


class TimestampsExtension(BlogExtension):
    """Cache entry timestamps.
    """
    
    entry_mixin = TimestampEntryMixin



#!/usr/bin/env python
"""
Module TITLE -- Simple Blog Title Extension
Sub-Package SIMPLEBLOG.EXTENSIONS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.stdlib.decotools import cached_property
from plib.stdlib.strings import universal_newline

from simpleblog.extensions import BlogExtension


class TitleEntryMixin(object):
    
    @cached_property
    def title_separator(self):
        return self.config.get('title_separator', universal_newline)
    
    def _do_load(self):
        raw = super(TitleEntryMixin, self)._do_load()
        if self.title_separator in raw:
            self._titlestr, raw = raw.split(self.title_separator, 1)
        else:
            self._titlestr = ""
        return raw


class TitleExtension(BlogExtension):
    """Parse entry title out of entry source.
    """
    
    entry_mixin = TitleEntryMixin
    
    def entry_get_title(self, entry):
        entry.load()
        return entry._titlestr



#!/usr/bin/env python
"""
Module CATEGORIES -- Simple Blog Categories Extension
Sub-Package SIMPLEBLOG.EXTENSIONS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import os

from plib.stdlib.decotools import cached_property
from plib.stdlib.ostools import subdirs

from simpleblog.extensions import (BlogExtension,
    NamedEntries, get_links)


class BlogCategory(NamedEntries):
    """Category of blog entries.
    """
    
    sourcetype = 'category'
    typename = "Category"
    
    @cached_property
    def prefix(self):
        return self.config.get('categories_prefix', "")
    
    def _get_entries(self):
        return [
            e for e in self.blog.all_entries if e.category == self.name
        ]


def make_category_link(category):
    return '<a href="/{0}/" title="Category">{0}</a>'.format(category)


class CategoryExtension(BlogExtension):
    """Add category to entry and category pages to blog.
    """
    
    def entry_post_init(self, entry):
        # The category will be an empty string for entries in
        # the root entries dir instead of a subdir
        entry.category, entry.name = os.path.split(entry.name)
        
        if entry.category:
            link = make_category_link(entry.category)
        else:
            link = "(None)"
        entry.metadata.update(
            categorylink=link
        )
    
    def entry_mod_attrs(self, entry, attrs, format, params):
        attrs.update(
            category=entry.category
        )
        return attrs
    
    def blog_mod_all_entries(self, blog, all_entries):
        # Entries can now live in subdirs of entries_dir (each
        # subdir defines a category)
        return all_entries + [
            blog.entry_class(blog, os.path.join(subdir, name))
            for subdir in subdirs(blog.entries_dir)
            for name in blog.filter_entries(os.path.join(blog.entries_dir, subdir))
        ]
    
    def blog_mod_sources(self, blog, sources):
        
        blog.category_names = set(subdirs(blog.entries_dir))
        
        blog.all_categories = [
            BlogCategory(blog, catname)
            for catname in blog.category_names
        ]
        
        blog.metadata['category_links'] = get_links(blog.all_categories)
        
        sources.extend(
            (category, "html")
            for category in blog.all_categories
        )
        
        return sources



#!/usr/bin/env python
"""
Module FEED -- Simple Blog Feed Extension
Sub-Package SIMPLEBLOG.EXTENSIONS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import re
import time as _time
from datetime import tzinfo, timedelta, datetime

from plib.stdlib.decotools import cached_property

from simpleblog import (extendable_property,
    html_newline,
    weekdaynames, monthnames)
from simpleblog.extensions import BlogExtension


# TZINFO code cribbed from Python docs for datetime module

ZERO = timedelta(0)
HOUR = timedelta(hours=1)


class UTC(tzinfo):
    
    def utcoffset(self, dt):
        return ZERO
    
    def tzname(self, dt):
        return "UTC"
    
    def dst(self, dt):
        return ZERO


STDOFFSET = timedelta(seconds = -_time.timezone)
if _time.daylight:
    DSTOFFSET = timedelta(seconds = -_time.altzone)
else:
    DSTOFFSET = STDOFFSET

DSTDIFF = DSTOFFSET - STDOFFSET


class LocalTimezone(tzinfo):

    def utcoffset(self, dt):
        if self._isdst(dt):
            return DSTOFFSET
        else:
            return STDOFFSET

    def dst(self, dt):
        if self._isdst(dt):
            return DSTDIFF
        else:
            return ZERO

    def tzname(self, dt):
        return _time.tzname[self._isdst(dt)]

    def _isdst(self, dt):
        tt = (dt.year, dt.month, dt.day,
              dt.hour, dt.minute, dt.second,
              dt.weekday(), 0, 0)
        stamp = _time.mktime(tt)
        tt = _time.localtime(stamp)
        return tt.tm_isdst > 0


tz_utc = UTC()

tz_local = LocalTimezone()


template_rss20 = "{0}, {1.day:02d} {2} {1.year} {1.hour:02d}:{1.minute:02d} GMT"

template_atom = "{0.year}-{0.month:02d}-{0.day:02d}T{0.hour:02d}:{0.minute:02d}:00Z"


class FeedEntryMixin(object):
    
    @cached_property
    def timestamp_utc(self):
        t = self.timestamp
        t_local = datetime(t.year, t.month, t.day, t.hour, t.minute, tzinfo=tz_local)
        return t_local.astimezone(tz_utc)
    
    @extendable_property()
    def timestamp_atom(self):
        t = self.timestamp_utc
        return template_atom.format(t)
    
    @extendable_property()
    def timestamp_rss20(self):
        t = self.timestamp_utc
        weekday = weekdaynames[t.weekday()]
        monthname = monthnames[t.month]
        return template_rss20.format(weekday, t, monthname)


class FeedBlogMixin(object):
    
    @extendable_property()
    def feed_formats(self):
        return set(["rss20", "atom"]).intersection(self.index_formats)


re_link = re.compile(r'<a href=\"\/([A-Za-z0-9\-\/\.]+)\"')


def fixup_relative_links(html, root_url):
    """Convert relative links in HTML to absolute links based on root URL.
    """
    return re.sub(
        re_link,
        lambda m: '<a href="{0}/{1}"'.format(root_url, m.group(1)),
        html
    )


feedlink_template_rss20 = """<link rel="alternate" type="application/rss+xml" title="RSS 2.0"
      href="{rss20_url}">"""

feedlink_template_atom = """<link rel="alternate" type="application/atom+xml" title="Atom"
      href="{atom_url}">"""


class FeedExtension(BlogExtension):
    """Add RSS and Atom feed links to home page.
    """
    
    entry_mixin = FeedEntryMixin
    blog_mixin = FeedBlogMixin
    
    def entry_mod_body(self, entry, body, format, params):
        if format in entry.blog.feed_formats:
            return fixup_relative_links(body, entry.blog.metadata['root_url'])
        return body
    
    def entry_mod_attrs(self, entry, attrs, format, params):
        do_rss20 = 'rss20' in entry.blog.feed_formats
        do_atom = 'atom' in entry.blog.feed_formats
        if do_rss20 or do_atom:
            attrs.update(
                root_url=entry.blog.metadata['root_url'],
                author=entry.blog.metadata['author'],
                email=entry.blog.metadata['email']
            )
            if do_rss20:
                attrs.update(
                    timestamp_rss20=entry.timestamp_rss20
                )
            if do_atom:
                attrs.update(
                    timestamp_atom=entry.timestamp_atom
                )
        return attrs
    
    def page_mod_attrs(self, page, attrs):
        if 'atom' in page.blog.feed_formats:
            attrs.update(
                sys_timestamp_atom=template_atom.format(datetime.utcnow())
            )
        if 'rss20' in page.blog.feed_formats:
            language = page.blog.metadata['language']
            try:
                country = page.blog.metadata['country'].lower()
            except KeyError:
                pass
            else:
                language = '-'.join([language, country])
            attrs.update(
                blog_language_rss20=language
            )
        return attrs
    
    def blog_mod_required_metadata(self, blog, data):
        if 'rss20' in blog.feed_formats:
            data.add('language')
        return data
    
    def blog_mod_default_metadata(self, blog, data):
        if 'rss20' in blog.feed_formats:
            data.update(
                rss20_url="/index.rss20"
            )
        if 'atom' in blog.index_formats:
            data.update(
                atom_url="/index.atom",
            )
        return data
    
    def blog_mod_sources(self, blog, sources):
        feedlinks = []
        if 'rss20' in blog.feed_formats:
            feedlinks.append(
                feedlink_template_rss20.format(**blog.metadata)
            )
        if 'atom' in blog.feed_formats:
            feedlinks.append(
                feedlink_template_atom.format(**blog.metadata)
            )
        blog.metadata['feed_links'] = html_newline.join(feedlinks)
        return sources



#!/usr/bin/env python
"""
Module RENDER_MARKDOWN -- Simple Blog Markdown Extension
Sub-Package SIMPLEBLOG.EXTENSIONS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from markdown import markdown

from simpleblog import html_newline
from simpleblog.extensions import BlogExtension


class MarkdownEntryMixin(object):
    
    def _do_render(self, rawdata):
        html = markdown(rawdata, output_format="html4")
        
        # Extra pretty formatting to match the behavior
        # of the Pyblosxom markdown formatter (indent
        # blockquotes and add blank lines between
        # paragraphs)
        lines = html.split(html_newline)
        out = []
        in_blockquote = False
        blankline_flag = False
        for line in lines:
            if blankline_flag:
                if not line.endswith("</blockquote>"):
                    out.append("")
                blankline_flag = False
            if in_blockquote:
                if line.endswith("</blockquote>"):
                    in_blockquote = False
                else:
                    line = "  {}".format(line)
            else:
                if line.startswith("<blockquote>"):
                    in_blockquote = True
            out.append(line)
            if (
                line.endswith("</p>") or
                line.endswith("</h1>") or
                line.endswith("</pre>") or
                line.endswith("</blockquote>")
            ):
                blankline_flag = True
        return html_newline.join(out)


class MarkdownExtension(BlogExtension):
    """Markdown parsing of entry source.
    """
    
    entry_mixin = MarkdownEntryMixin



#!/usr/bin/env python
"""
Sub-Package SIMPLEBLOG.COMMANDS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""


class BlogCommand(object):
    """Base class for command mechanism.
    """
    
    options = None
    arguments = None
    description = None
    epilog = None
    
    def __init__(self, opts, args):
        self.opts = opts
        self.args = args
    
    def run(self, config, blog):
        raise NotImplementedError



#!/usr/bin/env python
"""
Package SIMPLEBLOG -- Simple Python blog system
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import os
from collections import defaultdict
from datetime import datetime
from functools import wraps
from operator import attrgetter

from yaml import load

from plib.stdlib.builtins import first
from plib.stdlib.decotools import (
    cached_method, cached_property, memoize_generator)
from plib.stdlib.iters import suffixed_items
from plib.stdlib.version import version_string


__version__ = (0, 1)


# CONFIG

class BlogError(Exception):
    pass

class BlogConfigError(BlogError):
    pass


class BlogConfig(object):
    """Blog configuration.
    """
    
    settings = {}
    
    def __init__(self, filename=None):
        self.filename = filename or "config.yaml"
        if os.path.isfile(self.filename):
            with open(self.filename, 'rU') as f:
                self.settings = load(f)
        
        # This will prove to be convenient
        self.get = self.settings.get
    
    def __getattr__(self, name):
        try:
            return self.settings[name]
        except KeyError:
            raise AttributeError


# BASE


class BlogObject(object):
    """Base object for items managed by blog.
    """
    
    def __init__(self, blog):
        self.blog = blog
        self.config = self.blog.config
    
    @cached_property
    def entries_dir(self):
        return self.config.get('entries_dir', "entries")
    
    @cached_property
    def entry_ext(self):
        return self.config.get('entry_ext', ".html")
    
    @cached_property
    def template_dir(self):
        return self.config.get('template_dir', "templates")
    
    @cached_method
    def template_file(self, kind, format):
        return os.path.join(self.template_dir, "{0}.{1}".format(kind, format))
    
    @cached_method
    def template_data(self, kind, format):
        with open(self.template_file(kind, format), 'rU') as f:
            return f.read()


extension_types = {}

extension_map = defaultdict(list)

noresult = object()  # marker for no result returned

noreturn = object()  # marker for no return value at all


def check_extensions(instance, etype, key,
                     args=None, kwargs=None, result=noresult, modifying=False):
    args = args or ()
    kwargs = kwargs or {}
    for extension in extension_map.get(etype, ()):
        ext = getattr(extension, key, None)
        if ext is not None:
            if result is noreturn:
                ext(instance, *args, **kwargs)
            elif modifying:
                result = ext(instance, result, *args, **kwargs)
            else:
                result = ext(instance, *args, **kwargs)
            if (not modifying) and (result not in (noresult, noreturn)):
                break
    return result


class extendable_attr(object):
    """Base class for attributes that will use the extension mechanism.
    
    An instance of this class is applied as a decorator to mark a
    property or method of an extendable class that will use the
    extension mechanism. The instance wraps the function that
    implements the base property or method when it is called as
    a decorator. Arguments to the decorator give decorators that
    will be applied to the fully implemented property or method.
    Note that even if there are no arguments, the decorator must
    be applied with a call, like so:
    
    @extendable_property()
    def prop(self):
        ...
    """
    
    attrtype = None
    
    def __init__(self, *decorators):
        self._decorators = decorators
    
    def __call__(self, func):
        # What happens when this class is called as a decorator
        self._func = func
        return self
    
    def make_method(self, etype, name):
        assert name == self._func.__name__
        getkey = '{0}_get_{1}'.format(etype, name)
        modkey = '{0}_mod_{1}'.format(etype, name)
        @wraps(self._func)
        def meth(innerself, *args, **kwargs):
            # Check for extensions that might override the original function
            result = check_extensions(innerself, etype, getkey, args, kwargs)
            # Only call original function if no extension overrode it
            if result is noresult:
                result = self._func(innerself, *args, **kwargs)
            # Now check for extensions that want to modify the result
            result = check_extensions(innerself, etype, modkey, args, kwargs,
                result, True)
            return result
        return meth
    
    def as_extended(self, cls, name):
        meth = self.make_method(cls.extension_type, name)
        for d in reversed(self._decorators):
            # Reversing makes the declaration order more intuitive
            meth = d(meth)
        return self.attrtype(meth)


class extendable_property(extendable_attr):
    """Property to be fully implemented by ``extendable`` class decorator.
    """
    
    attrtype = cached_property


class extendable_method(extendable_attr):
    """Method to be fully implemented by ``extendable`` class decorator.
    """
    
    attrtype = cached_method


def extend_attributes(cls):
    """Implement all extendable attributes in cls.
    """
    
    for key in dir(cls):
        attr = getattr(cls, key)
        if isinstance(attr, extendable_attr):
            setattr(cls, key, attr.as_extended(cls, key))


def extendable(cls):
    """Class decorator to implement extendable properties.
    """
    
    # Determine extension type and add to map
    etype = cls.__name__.lower()
    eprefix = 'blog'
    if len(etype) > len(eprefix):
        etype = etype[len(eprefix):]
    cls.extension_type = etype
    extension_types[etype] = cls
    
    # Implement all extendable attributes
    extend_attributes(cls)
    
    # Implement post_init call at end of constructor
    initkey = '{}_post_init'.format(etype)
    oldinit = cls.__init__
    @wraps(oldinit)
    def _newinit(self, *args, **kwargs):
        oldinit(self, *args, **kwargs)
        check_extensions(self, etype, initkey, result=noreturn)
    cls.__init__ = _newinit
    
    return cls


# ENTRY

weekdaynames = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]

monthnames = [None,
    "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]

html_newline = '\n'  # FIXME find out markdown's newline spec


@extendable
class BlogEntry(BlogObject):
    """Single blog entry.
    """
    
    sourcetype = 'entry'
    multisource = None
    
    def __init__(self, blog, name):
        BlogObject.__init__(self, blog)
        self.cachekey = self.name = name
        self.urlpath = "/{}".format(name)
        self.heading = "Single Entry"
        self.metadata = {}
    
    @extendable_property()
    def title(self):
        return self.name
    
    @extendable_property()
    def filename(self):
        return os.path.join(self.entries_dir, self.cachekey + self.entry_ext)
    
    @extendable_property()
    def timestamp(self):
        return datetime.fromtimestamp(os.path.getmtime(self.filename))
    
    @cached_property
    def timestamp_vars(self):
        t = self.timestamp
        return dict(
            year=t.year,
            month=t.month,
            day=t.day,
            hour=t.hour,
            minute=t.minute,
            second=t.second,
            weekdayname=weekdaynames[t.weekday()],
            monthname=monthnames[t.month]
        )
    
    @cached_property
    def timestamp_template(self):
        return self.config.get('timestamp_template',
            "{hour:02d}:{minute:02d}")
    
    @extendable_property()
    def timestamp_formatted(self):
        return self.timestamp_template.format(**self.timestamp_vars)
    
    @cached_property
    def datestamp_template(self):
        return self.config.get('datestamp_template',
            "{weekdayname}, {day:02d} {monthname} {year}")
    
    @extendable_property()
    def datestamp_formatted(self):
        return self.datestamp_template.format(**self.timestamp_vars)
    
    # Note that source is *not* an extendable property! It must
    # always be the actual source loaded from the filesystem,
    # or whatever other source is being used. Mixins should
    # override _get_source instead, to ensure that the caching
    # mechanism works properly.
    
    @cached_property
    def source(self):
        return self._get_source()
    
    def _get_source(self):
        with open(self.filename, 'rU') as f:
            return f.read()
    
    # Load is also *not* extendable, because it must be callable
    # from any code that needs to ensure that the actual data is
    # loaded. Mixins should override _do_load.
    
    @cached_method
    def load(self):
        """Load raw data.
        """
        return self._do_load()
    
    def _do_load(self):
        return self.source
    
    # Render is not extendable because, like load, it needs to be
    # callable from any code that needs to render raw data. Mixins
    # should override _do_render.
    
    @cached_method
    def render(self, rawdata):
        """Convert raw data into rendered data.
        """
        return self._do_render(rawdata)
    
    def _do_render(self, rawdata):
        return rawdata
    
    @extendable_property()
    def rendered(self):
        return self.render(self.load())
    
    @extendable_method()
    def body(self, format, params):
        return self.rendered
    
    @cached_method
    def make_permalink(self, format):
        return "{}.{}".format(self.urlpath, format)
    
    @extendable_method()
    def attrs(self, format, params):
        return dict(
            name=self.name,
            cachekey=self.cachekey,
            urlpath=self.urlpath,
            title=self.title,
            body=self.body(format, params),
            year=self.timestamp.year,
            month=self.timestamp.month,
            day=self.timestamp.day,
            hour=self.timestamp.hour,
            minute=self.timestamp.minute,
            timestamp=self.timestamp_formatted,
            datestamp=self.datestamp_formatted,
            permalink=self.make_permalink(format),
            **self.metadata
        )
    
    @extendable_method()
    def template(self, format):
        return self.template_data("entry", format)
    
    @extendable_method()
    def formatted(self, format, params):
        return self.template(format).format(**self.attrs(format, params))


# CONTAINERS


class BlogEntries(BlogObject):
    """Container for a set of blog entries.
    """
    
    sourcetype = None
    multisource = None
    urlshort = None
    
    def __init__(self, blog):
        BlogObject.__init__(self, blog)
    
    def _get_entries(self):
        raise NotImplementedError
    
    @cached_property
    def entry_sort_key(self):
        return self.config.get('entry_sort_key', 'timestamp')
    
    @cached_property
    def entry_sort_reversed(self):
        return self.config.get('entry_sort_reversed', True)
    
    @cached_property
    def entries(self):
        return sorted(self._get_entries(),
            key=attrgetter(self.entry_sort_key),
            reverse=self.entry_sort_reversed
        )
    
    def _get_urlpath(self):
        if self.urlshort is not None:
            return "{}index".format(self.urlshort)
        raise NotImplementedError
    
    @cached_property
    def urlpath(self):
        return self._get_urlpath()


class BlogIndex(BlogEntries):
    
    sourcetype = 'blog'
    title = "Home"
    heading = "Home Page"
    urlshort = "/"
    
    def _get_entries(self):
        return self.blog.all_entries


# PAGES


def prefixed_keys(mapping, prefix):
    """Return dict with all keys prefixed by ``prefix``.
    """
    tmpl = prefix + "{}"
    return mapping.__class__(
        (tmpl.format(key), mapping[key])
        for key in mapping
    )


class BlogEntryParams(object):
    """Convenience object to store format parameters.
    """
    
    def __init__(self, mapping=None, **kwargs):
        self.update(mapping, **kwargs)
    
    def update(self, mapping=None, **kwargs):
        for obj in (mapping, kwargs):
            if obj:
                for k, v in obj.iteritems():
                    setattr(self, k, v)


@extendable
class BlogPage(BlogObject):
    """Single page containing one or more entries.
    """
    
    def __init__(self, blog, source, format, pagenum=None):
        BlogObject.__init__(self, blog)
        self.source = source
        if isinstance(source, BlogEntry):
            self.entries = [source]
        else:
            self.entries = source.entries
        self.format = format
        self.pagenum = pagenum
        if self.pagenum and (self.pagenum > 1):
            pagemarker = self.pagenum
        else:
            pagemarker = ""
        self.urlpath = "{0}{1}.{2}".format(
            source.urlpath, pagemarker, format)
    
    # The format_entries generator is not extendable;
    # mixins can override _get_format_entries
    
    @memoize_generator
    def format_entries(self):
        return self._get_format_entries()
    
    @extendable_method()
    def entry_params(self, entry):
        return BlogEntryParams()
    
    def _get_format_entries(self):
        for entry in self.entries:
            yield entry.formatted(self.format, self.entry_params(entry))
    
    @cached_property
    def no_entries(self):
        return self.config.get('no_entries_content', "<p>No entries found!</p>")
    
    @extendable_property()
    def body(self):
        if self.entries:
            return html_newline.join(self.format_entries())
        return self.no_entries
    
    @extendable_property()
    def template(self):
        return self.template_data("page", self.format)
    
    @extendable_property()
    def attrs(self):
        metadata = dict(
            title=self.source.title,
            heading=self.source.heading,
            entries=self.body
        )
        return dict(
            prefixed_keys(self.blog.metadata, 'blog_'),
            sys_gen_name='simpleblog',
            sys_gen_uri="http://pypi.python.org/pypi/simpleblog",
            sys_gen_version=version_string(__version__),
            **prefixed_keys(metadata, 'page_')
        )
    
    @extendable_property()
    def formatted(self):
        return self.template.format(**self.attrs)


# BLOG


class BlogMetadataError(BlogError):
    pass


@extendable
class Blog(BlogObject):
    """The entire blog.
    """
    
    def __init__(self, config, filename=None):
        self.blog = self
        self.config = config
        self.filename = filename or "blog.yaml"
        self.metadata = {}
        if os.path.isfile(self.filename):
            with open(self.filename, 'rU') as f:
                self.metadata.update(load(f))
        for key in self.required_metadata:
            if key not in self.metadata:
                raise BlogMetadataError("{} missing from blog metadata".format(key))
        for key, value in self.default_metadata.iteritems():
            self.metadata.setdefault(key, value.format(**self.metadata))
    
    @extendable_property()
    def required_metadata(self):
        return set()
    
    @extendable_property()
    def default_metadata(self):
        return {}
    
    @cached_property
    def entry_class(self):
        return extension_types['entry']
    
    @cached_method
    def filter_entries(self, path):
        return suffixed_items(os.listdir(path), self.entry_ext)
    
    @extendable_property()
    def all_entries(self):
        return [
            self.entry_class(self, name)
            for name in self.filter_entries(self.entries_dir)
        ]
    
    @cached_property
    def page_class(self):
        return extension_types['page']
    
    @extendable_property()
    def index_formats(self):
        return set(self.config.get('index_formats', ["html"]))
    
    @extendable_property()
    def entry_formats(self):
        return set(self.config.get('entry_formats', ["html"]))
    
    @extendable_property()
    def sources(self):
        
        self.index_entries = BlogIndex(self)
        
        return [
            (self.index_entries, format)
            for format in self.index_formats
        ] + [
            (entry, format)
            for entry in self.all_entries
            for format in self.entry_formats
        ]
    
    @extendable_property()
    def pages(self):
        return [
            self.page_class(self, source, format)
            for source, format in self.sources
        ]


# COMMANDS


def first_subclass(o, c):
    """Return first object in o that is a subclass of c
    """
    return first(
        obj for obj in vars(o).itervalues()
        if (obj is not c)
        and isinstance(obj, type)
        and issubclass(obj, c)
    )


def load_sub(subtype, name, err, subcls):
    # FIXME allow subtype objects to "live" in other places as well
    from importlib import import_module
    try:
        mod = import_module("simpleblog.{0}s.{1}".format(subtype, name))
    except ImportError:
        raise err("{0} {1} not found!".format(subtype, name))
    klass = first_subclass(mod, subcls)
    if not klass:
        raise err("no {0} in {1} module!".format(subtype, name))
    return mod, klass


class BlogExtensionError(BlogError):
    pass


def initialize(config):
    extensions = config.get("extensions", ())
    if extensions:
        from simpleblog.extensions import BlogExtension
        # This hack is useful for extensions that need to access the
        # config in module or class level code, e.g., in decorators
        BlogExtension.config = config
        for extname in extensions:
            mod, klass = load_sub("extension", extname,
                BlogExtensionError, BlogExtension)
            mod.extension = klass(config)


def load_blog(opts):
    config = BlogConfig(opts.configfile)
    initialize(config)
    blog = extension_types['blog'](config, opts.blogfile)
    return config, blog


class BlogCommandError(BlogError):
    pass


global_optlist = (
    ("-c", "--configfile",
        { 'help': "the configuration file name (defaults to config.yaml)" }
    ),
    ("-b", "--blogfile",
        { 'help': "the blog metadata file name (defaults to blog.yaml)" }
    )
)

global_arglist = (
    ("command", { 'nargs': "?", 'default': "" }),
)


def run(cmdname, result=None, remaining=None):
    from simpleblog.commands import BlogCommand
    mod, klass = load_sub("command", cmdname,
        BlogCommandError, BlogCommand)
    
    optlist = klass.options or ()
    arglist = klass.arguments or ()
    description = klass.description
    epilog = klass.epilog
    if optlist or arglist or remaining:
        from plib.stdlib.options import parse_options
        opts, args = parse_options(
            optlist, arglist, description, epilog, remaining, result
        )
    else:
        from plib.stdlib.options import prepare_specs, make_objs
        optlist, arglist = prepare_specs(global_optlist, global_arglist)
        opts, args = make_objs(result, optlist, arglist)
    
    cmd = klass(opts, args)
    config, blog = load_blog(opts)
    cmd.run(config, blog)



#!/usr/bin/env python
"""
Module TAGS -- Simple Blog Tags Extension
Sub-Package SIMPLEBLOG.EXTENSIONS
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.stdlib.decotools import cached_property, convert
from plib.stdlib.strings import split_string, universal_newline

from simpleblog import extendable_property, html_newline
from simpleblog.caching import cached
from simpleblog.extensions import (BlogExtension,
    NamedEntries, get_links)


tags_file = BlogExtension.config.get('tags_file', "tags")


def makelink(name, prefix=None):
    if prefix:
        href = "{0}/{1}".format(prefix, name)
    else:
        href = name
    return '<a href="/{0}/">{1}</a>'.format(href, name)


class Tagset(object):
    """Blog entry tag set.
    
    Wraps a set of blog tags to customize for usage here. The
    ``__str__`` method ensures that the tags can be written to
    and read back from the tag cache file, as well as from the
    tag string in the entry source file. Other important methods
    are delegated to the underlying set object.
    """
    
    def __init__(self, s):
        self._tags = frozenset(t.strip() for t in s.split(','))
    
    def __str__(self):
        return ','.join(sorted(self._tags))
    
    def __iter__(self):
        return iter(self._tags)
    
    def __contains__(self, v):
        return v in self._tags
    
    def __len__(self):
        return len(self._tags)


class BlogTag(NamedEntries):
    """Entries matching a given tag.
    """
    
    sourcetype = 'tag'
    multisource = 'tags'
    typename = "Tag"
    
    @cached_property
    def prefix(self):
        return self.config.get('tags_prefix', "")
    
    def _get_entries(self):
        return [
            e for e in self.blog.all_entries if self.name in e.tags
        ]


class TagsEntryMixin(object):
    
    @extendable_property()
    def tags_prefix(self):
        return self.config.get('tags_prefix', "#tags ")
    
    @extendable_property()
    def tags_suffix(self):
        return self.config.get('tags_suffix', universal_newline)
    
    def _do_load(self):
        raw = super(TagsEntryMixin, self)._do_load()
        pre, mid, post = split_string(raw,
            self.tags_prefix, self.tags_suffix, find_newlines=False)
        if pre and post:
            self._tagstr = mid
            raw = ''.join([pre[:-len(self.tags_prefix)], post[len(self.tags_suffix):]])
        else:
            self._tagstr = ""
        return raw
    
    @extendable_property(
        cached(tags_file, objtype=Tagset)
    )
    def tags(self):
        self.load()
        return self._tagstr


class TagsExtension(BlogExtension):
    """Add tags to entry and tag pages to blog.
    """
    
    entry_mixin = TagsEntryMixin
    
    @cached_property
    def tags_prefix(self):
        return self.config.get('tags_prefix', "")
    
    def entry_post_init(self, entry):
        entry.metadata.update(
            taglinks=',{}'.format(html_newline).join(
                makelink(tag, self.tags_prefix) for tag in sorted(entry.tags)
            )
        )
    
    def entry_mod_attrs(self, entry, attrs, format, params):
        attrs.update(
            tags=str(entry.tags)
        )
        return attrs
    
    def blog_mod_sources(self, blog, sources):
        
        blog.tag_names = set(t
            for e in blog.all_entries
            for t in e.tags
        )
        
        blog.all_tags = [
            BlogTag(blog, tagname)
            for tagname in blog.tag_names
        ]
        
        blog.metadata.update(
            tag_links=get_links(blog.all_tags)
        )
        
        sources.extend(
            (tag, "html")
            for tag in blog.all_tags
        )
        
        return sources



#!/usr/bin/env python -u
"""
Setup script for SIMPLEBLOG package
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from datetime import datetime
from itertools import islice, takewhile

from distutils.core import setup


version = "0.1"

h3_prefix = '### '


def h3_underline(line, underline='~'):
    if line.endswith('\n'):
        line = line[:-1]
    return "{}\n".format(underline * len(line))


def make_readme(filename="README.md"):
    with open(filename, 'rU') as f:
        lines = f.readlines()
    outlines = [
        "**README** for SIMPLEBLOG {}\n".format(version),
        "\n",
        ":Author:        Peter A. Donis\n",
        ":Release Date:  {}\n".format(datetime.now().strftime("%d %b %Y"))
    ]
    for line in islice(lines, 2, None):
        is_h3line = line.startswith(h3_prefix)
        if is_h3line:
            line = line[len(h3_prefix):]
        outlines.append(line)
        if is_h3line:
            outlines.append(h3_underline(line))
    with open("README", 'w') as f:
        f.writelines(outlines)


def long_description(filename="README"):
    with open(filename, 'rU') as f:
        lines = f.readlines()
    return "".join(takewhile(
        lambda line: not line.startswith("Copyright and License"),
        islice(lines, 5, None)
    ))


make_readme()

setup(
    name="simpleblog",
    version=version,
    description="A simple Python blogging system.",
    long_description=long_description(),
    author="Peter A. Donis",
    author_email="peterdonis@alum.mit.edu",
    url="http://pypi.python.org/pypi/simpleblog",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python"
    ],
    requires=[
        'plib (>=0.8)',
        'yaml'
    ],
    packages=[
        'simpleblog',
        'simpleblog.commands',
        'simpleblog.extensions'
    ],
    scripts=[
        'scripts/simpleblog-run'
    ]
)



#!/usr/bin/env python
"""
Module CACHING -- Simple Blog Caching Mechanism
Package SIMPLEBLOG
Copyright (C) 2012 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import os
from functools import wraps

from plib.stdlib.decotools import cached_property

from simpleblog import BlogObject


class BlogCache(BlogObject):
    """Cache for data associated with blog entries.
    
    This allows the blog structure and metadata to be assembled
    without having to stat or open and load any entry files.
    """
    
    def __init__(self, blog, cachename, reverse=False, objtype=None):
        BlogObject.__init__(self, blog)
        self.cachename = cachename
        self.reverse = reverse
        self.objtype = objtype
    
    @cached_property
    def cache_dir(self):
        return self.config.get('cache_dir', self.entries_dir)
    
    @cached_property
    def filename(self):
        return os.path.join(self.cache_dir, self.cachename)
    
    @cached_property
    def cache(self):
        with open(self.filename, 'rU') as f:
            lines = f.readlines()
        if self.reverse:
            def r(t):
                return tuple(reversed(t))
        else:
            def r(t):
                return t
        if self.objtype:
            def e(t):
                return t[0], self.objtype(t[1])
        else:
            def e(t):
                return t
        return dict(
            e(r(line.split())) for line in lines
        )
    
    def save(self):
        items = self.cache.iteritems()
        if self.reverse:
            items = ((v, k) for k, v in items)
        lines = sorted("{0!s} {1!s}\n".format(a, b) for a, b in items)
        with open(self.filename, 'w') as f:
            f.writelines(lines)


cache_map = {}


def cached(cachename, reverse=False, objtype=None):
    """Decorator for blog entry properties that should be cached.
    """
    
    def decorator(f):
        @wraps(f)
        def fcache(self, *args, **kwargs):
            try:
                cacheobj = cache_map[cachename]
            except KeyError:
                cacheobj = cache_map[cachename] = BlogCache(
                    self.blog, cachename, reverse, objtype
                )
            cache = cacheobj.cache
            try:
                return cache[self.cachekey]
            except KeyError:
                value = f(self, *args, **kwargs)
                if cacheobj.objtype is not None:
                    value = cacheobj.objtype(value)
                cache[self.cachekey] = value
                cacheobj.save()
                return value
        return fcache
    return decorator
